"""
NOTE: this file contains legacy code which was used originally with Quantum Inspire (QI) 1.0,
and has now been translated in order to be used with QI 2.0.
Apart from the translation to QI 2.0, this code and format are not being currently supported.
"""

import os
import numpy as np
from datetime import datetime
import pandas as pd
from qiskit_quantuminspire.qi_provider import QIProvider

def prepare_file(basename: str="",
                 suffix: str="",
                 doraw: int=0):
    """
    Creates the file name according to the basename and suffix that you provide.
    """

    histname="hist_"+basename
    circuitname="circuit_"+basename
    rawname="Raw_"+basename
    if (len(suffix)>0):
        histname+='_'+suffix
        circuitname+='_'+suffix
        rawname+='_'+suffix
    histname+='_API.txt'
    circuitname+='_API.pdf'
    rawname+="_API"

    file=open(histname,'w')
    file=open(circuitname,'w')
    
    if (doraw==0):
        return histname, circuitname
    else:
        return histname, circuitname, rawname

def GetTimeStamp():
    """
    Returns the timestamp of the current date and time
    """

    current_date = datetime.datetime.now()
    thisyear=str(current_date.year)
    thismonth="0"+str(current_date.month)
    thisday="0"+str(current_date.day)
    thishour="0"+str(current_date.hour)
    thisminute="0"+str(current_date.minute)
    thissecond="0"+str(current_date.second)
    timestamp=thisyear[-2:]+thismonth[-2:]+thisday[-2:]+"_"+thishour[-2:]+thisminute[-2:]+thissecond[-2:]
    return timestamp

def create_new_data_folder(datadir: str):
    """
    Crates the folder with the current date in the specified path.
    """

    data_folder_path = datadir + "/Data"
    try:
        os.makedirs(data_folder_path, exist_ok=False)
    except:
        print("Data folder already exists")

    now = datetime.now()
    timestamp = now.strftime("%Y%m%d")
    today_path = data_folder_path + f"/{timestamp}"
    try:
        os.makedirs(today_path, exist_ok=False)
    except:
        print(f"Folder with timestamp {timestamp} already exists")
    
    os.chdir(today_path)  # change the current working directory to the specified path
    return today_path

def api_run_and_save(param,
                     Qcircuit,
                     histname: str="hist.txt",
                     circuit_name: str="cqasm.txt",
                     shots: int=16384,
                     backend_name: str='Tuna-9',
                     get_results: bool=True,
                     get_hist_data: bool=False,
                     measurement_list: list=[],
                     get_raw_data: bool=False,
                     rawdata_filename: str="rawdata",
                     timeout: int = 1200):
    """
    Runs QI with qiskit program and returns histogram and the raw data
    A copy of the cqasm program is saved to file circuit_name.

    param:              a reference number that you are free to choose.
    Qcircuit:           Qiskit quantum circuit.
    histname:           file name where you want to save the histogram data.                        
    circuit_name:       name of the file in which you want to save the quantum circuit.
    shots:              desired number of shots. For Tuna-9, the max is 16384.
    backend_name:       specify the name of the backend that you want to use.
    get_results:        False: do not return the measurement result
                        True: return the measurement result
    get_hist_data:      False: do not return the histogram data
                        True: return the histogram data (if this is True make sure to specify the measurement_list)
    measurement_list:   each entry of the list is equal to the number of measurements done simultaneously in the algorithm.
                        e.g. measurement_list = [4, 2, 1], it means that the rightmost entry of the classical bit string is the result of a single measurement,
                        the second and the third entries of the classical bit string are measured together and the last four are measured together.                        
    get_raw_data:       False: do not return the raw data
                        True: return the raw data
    rawdata_filename:   name of the raw data file you want to save
    """
    
    # Set the backend
    provider = QIProvider()
    backend = provider.get_backend(name = backend_name)
    
    Qcircuit.draw('mpl', filename = circuit_name)

    # Run the job
    job = backend.run(Qcircuit, shots = shots, memory = get_raw_data)
    results = job.result(timeout = timeout) # get the results

    # Get and save the histogram data
    if get_hist_data:
        histogram_data = results.get_counts()
        histogram_keys = dict()

        for entry_index, entry in enumerate(histogram_data):
            additional_len = Qcircuit.num_clbits - len(entry)
            for i in range(additional_len):
                entry = ('0' + (entry))
            histogram_keys[entry_index] = 'd' + entry

        data = np.column_stack([list(histogram_keys.values()), list(histogram_data.values())])

        process_data_and_save(data, measurement_list, histname)

    # Get and save the raw data (if asked)
    if get_raw_data:
        raw_data = results.get_memory()

        for entry in range(len(raw_data)):
            additional_len = Qcircuit.num_clbits - len(raw_data[entry])
            for i in range(additional_len):
                raw_data[entry] = '0' + raw_data[entry]

        for nr_shots in range(len(raw_data)):
            raw_data[nr_shots] = 'd'+str(raw_data[nr_shots])

        df = pd.DataFrame({ 
                                "Raw data values": raw_data
                            })
        # Save to a csv file
        output_file_rawdata = rawdata_filename+"_"+str(param)+".csv"
        df.to_csv(output_file_rawdata, index = False)

    if get_results:
        return results

def process_data_and_save(data, q, filename):
    q_cumsum = np.cumsum(q)  # Cumulative sum of q to determine slicing indices
    column_dicts = [{} for _ in q]  # Create an empty dictionary for each entry in q

    for row in data:
        bitstring = row[0][1:]  # Remove the 'd' at the start
        count = int(row[1])    # Count of the bitstring
        
        for i, length in enumerate(q):
            if i == 0:
                sliced = bitstring[:length]  # First slice
            else:
                sliced = bitstring[q_cumsum[i-1]:q_cumsum[i]]  # Subsequent slices
            
            # Populate the dictionary for this column
            if sliced in column_dicts[i]:
                column_dicts[i][sliced] += count  # Increment count if key exists
            else:
                column_dicts[i][sliced] = count  # Initialize key with count
    
    column_dicts_reversed = column_dicts[::-1]

    # Write results to a file
    with open(filename, 'w') as file:
        for i, col_dict in enumerate(column_dicts_reversed):
            file.write(f"{i}:")
            file.write(r"{")
            for key, value in col_dict.items():
                file.write(f"'{key}': {value}")
            file.write(r"}")
            file.write("\n")  # Add a blank line between dictionaries


"""
Utility functions for constructing and modifying specific parts
of an existing Qiskit QuantumCircuit object ('qc').

This module contains helper routines for state initialization,
measurement basis rotation, and readout circuit composition.

Authors: Marios Samiotis
"""

import numpy as np
from qiskit import QuantumCircuit, ClassicalRegister

def prepare_initial_state(qc: QuantumCircuit,
                          initial_state: str):
    """
    This function applies an initialization to an existing
    QuantumCircuit object. The initialization works only for either
    the state |0> or |1>.
    
    Args:
        qc (QuantumCircuit):
            The quantum circuit object.

        initial_state (str):
            A bitstring specifying the initial state. The order in the bitstring
            is 'qn-1,qn-2,...,q2,q1,q0'. The string can only contain '0' or '1'.
            e.g. for a quantum circuit with 3 qubits, initial_state = '100' will
            initialize qubits q0 and q1 in the state |0>, and qubit q2 in the state
            |1>.
    """
    
    nr_qubits = qc.num_qubits
    qc.barrier()
    for idx in range(nr_qubits):
        qc.reset(idx)
    qc.barrier()

    if len(initial_state) != nr_qubits:
        raise ValueError('Initial state must have same number of qubits defined.')
    for idx in range(len(initial_state)):
        if initial_state[idx] == '1':
            qc.x((nr_qubits-1) - idx)
    qc.barrier()

    return qc

def apply_pre_measurement_rotations(qc: QuantumCircuit,
                                    observable: str,
                                    bit_register: list = None):
    """
    This function applies all necessary pre-rotations prior to the measurement
    of an observable. It is meant to be applied on a QuantumCircuit object which
    does not contain any measurement blocks at its end.

    Args:
        qc (QuantumCircuit):
            The quantum circuit object.

        observable (str):
            The observable to be measured, expressed strictly in the Pauli basis,
            for which the pre-measurement rotations are being applied for.
            For an n-qubit Pauli string P, where Pauli Pi acts on qubit qi, the order
            in the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit observable string 'YX', qubit q0 is measured
            in the X basis, while qubit q1 is measured in the Y basis.

        bit_register (list):
            A list specifying the bit register for which the measurements outcomes
            will be stored in.
    """
    
    nr_qubits = len(observable)
    for idx in range(nr_qubits):
        if observable[idx] == 'I':
            return qc
        
        elif observable[idx] == 'X':
            qc.ry(-np.pi/2,(nr_qubits-1)-idx)
        elif observable[idx] == 'Y':
            qc.rx(np.pi/2,(nr_qubits-1)-idx)
        elif observable[idx] == 'Z':
            pass

        if bit_register is not None:
            qc.measure((nr_qubits-1)-idx, bit_register[(nr_qubits-1)-idx])
        else:
            qc.measure((nr_qubits-1)-idx,(nr_qubits-1)-idx)

    return qc

def apply_readout_circuit(qc: QuantumCircuit,
                          qubit_list: list):
    """
    This function appends a QuantumCircuit object with a quantum circuit
    containing mid-circuit measurements which are used in post-processing
    for constructing the readout assignment matrix. The matrix is later on
    applied on the raw measurement outcomes in order to mitigate readout
    errors.

    Args:
        qc (QuantumCircuit):
            The quantum circuit object.

        qubit_list (list):
            An ordered list specifying the qubits on which we apply the
            readout circuit on.
            e.g. qubit_list = [0, 2] will apply the readout circuit on
            qubits q0 and q2.
    """

    nr_qubits = len(qubit_list)
    readout_circuit = QuantumCircuit(qc.num_qubits, qc.num_clbits + nr_qubits * 2**nr_qubits, name=qc.name)

    binary_list = []
    for binary_str_idx in range(2**nr_qubits):
        binary_list.append(np.binary_repr(binary_str_idx, nr_qubits))

    for binary_str_idx in range(len(binary_list)):

        reversed_binary_string = binary_list[binary_str_idx][::-1]

        for qubit_idx in qubit_list:
            readout_circuit.reset(qubit_idx)

        for idx in range(len(reversed_binary_string)):
            if reversed_binary_string[idx] == '1':
                readout_circuit.x(qubit_list[idx])
        readout_circuit.barrier()

        readout_circuit.measure(qubit_list,
                                list(np.arange(start=qc.num_clbits + binary_str_idx*nr_qubits,
                                               stop=qc.num_clbits + (binary_str_idx+1)*nr_qubits, step=1)))
        readout_circuit.barrier()

    qc.barrier()
    additional_bits = ClassicalRegister(nr_qubits * 2**nr_qubits)
    qc.add_bits(additional_bits)
    return readout_circuit.compose(qc, front=True)


"""
Utility classes for storing and retrieving a job (project)
data and metadata in a structured manner in the local user directory.

Authors: Marios Samiotis
"""

import numpy as np
import json
import h5py
import warnings
import matplotlib.pyplot as plt
from pathlib import Path
from PIL import ImageFilter
from qiskit import qasm3
from qiskit_quantuminspire.cqasm import dumps
from qiskit_quantuminspire.qi_jobs import QIJob

class StoreProjectRecord:
    """
    This class is responsible for storing a job (project) record
    within the user's local 'Documents' directory.

    'Record' encapsulates both 'data', such as the measurement counts
    or even raw data shots, but also the 'metadata', such as information
    regarding the backend, the quantum circuits executed, etc.

    It works with both the Tuna backends but also the 'QX emulator' backend.
    """

    def __init__(self,
                 job: QIJob):
        """
        Args:
            job (QIJob):
                The user already-submitted job object, more correctly referred to as
                'project'. A project can contain multiple jobs, but for simplification,
                and also for legacy reasons, we keep referring to it as the 'job'.
        """

        self.create_project_directory(job)
        self.obtain_backend_metadata(job)
        self.store_project_json()
        for job_idx in range(len(job.circuits_run_data)):
            self.create_job_directory(job, job_idx)
            self.store_circuit_metadata(job, job_idx)
            self.store_job_result(job, job_idx)
            if self.raw_data_memory == True:
                self.store_raw_data(job, job_idx)

        return print(f"Successfully stored project record in the following directory:\n{str(self.project_dir)}\n")

    def create_project_directory(self,
                                 job: QIJob):
        """
        This instance method creates a new project folder within the local user
        Documents / QuantumInspireProjects directory. If this directory does not
        already exist, it will be created automatically.

        Args:
            job (QIJob):
                The user already-submitted job (project) object.
        """

        timestamp_utc = job.circuits_run_data[0].results.created_on # actually when the job finished, not when created
        timestamp = timestamp_utc.astimezone()
        self.date_timestamp = timestamp.strftime("%Y%m%d")
        self.job_0_timestamp = timestamp.strftime("%H%M%S")

        self.project_name = job.program_name
        self.project_dir = (
            Path.home() / "Documents" / "QuantumInspireProjects" / self.date_timestamp
            / f"{self.job_0_timestamp}_{self.project_name}"
        )
        self.project_dir.mkdir(parents=True, exist_ok=True)

    def obtain_backend_metadata(self,
                                job: QIJob):
        """
        This instance method retrieves and then stores all relevant metadata
        for the backend which was used to run the job(s) on.

        Args:
            job (QIJob):
                The user already-submitted job (project) object.
        """

        self.backend_name = job.backend().name
        self.backend_nr_qubits = job.backend().num_qubits
        self.backend_operations = []
        for entry in range(len(job.backend().operations)):
            self.backend_operations.append(str(job.backend().operations[entry]))
        self.backend_max_shots = job.backend().max_shots

        try: # since the user may have used an emulator
            figure = job.backend().coupling_map.draw()
            image = figure.resize((800, 800))
            sharpened = image.filter(ImageFilter.SHARPEN)
            image_array = np.array(sharpened)

            file_path = (
                Path(self.project_dir)
                / f"backend_coupling_map_{self.date_timestamp}_{self.job_0_timestamp}.png"
            )

            plt.clf()
            plt.imshow(image_array)
            plt.title(f'\n{self.date_timestamp}_{self.job_0_timestamp}\n{job.backend().name} coupling map\n', fontsize=18)
            plt.axis('off')
            plt.savefig(file_path, bbox_inches='tight', pad_inches=0)
            plt.close()
        except Exception as e:
            exception_message = str(e)
            warnings.warn(f"\nFailed storing backend {self.backend_name} coupling map figure.\nError message: {exception_message}\n",
                          UserWarning)

    def store_project_json(self):
        """
        This instance method stores job (project) related metadata within
        the project directory in a JSON format.
        """
        
        general_dict = {}
        general_dict['Project name'] = self.project_name
        general_dict['Project timestamp'] = f"{self.date_timestamp}_{self.job_0_timestamp}"
        general_dict['Backend name'] = self.backend_name
        general_dict['Backend number of qubits'] = self.backend_nr_qubits
        general_dict['Backend operations set'] = self.backend_operations
        general_dict['Backend maximum allowed shots'] = self.backend_max_shots

        file_path = (
            Path(self.project_dir)
            / f"project_metadata_{self.date_timestamp}_{self.job_0_timestamp}.json"
        )
        with open(file_path, 'w') as file:
            json.dump(general_dict, file, indent=3)

    def create_job_directory(self,
                             job: QIJob,
                             job_idx: int):
        """
        This instance method creates new directories for each job contained
        within the project.

        Args:
            job (QIJob):
                The user already-submitted job (project) object.

            job_idx (int):
                The job index for all jobs contained within the project.
                While a project may contain a certain number of jobs, e.g. N,
                it is generally true that the execution of all these jobs
                in the Quantum Inspire platform is not sequential with respect
                to the order with which those jobs were created.
                Therefore, job_idx is being utilized for clarity when storing
                the data, so that it follows the sequence with which the jobs were
                created.
        """
        
        timestamp_utc = job.circuits_run_data[job_idx].results.created_on # actually when the job finished, not when created
        timestamp = timestamp_utc.astimezone()
        self.date_timestamp = timestamp.strftime("%Y%m%d")
        self.job_timestamp = timestamp.strftime("%H%M%S")
        self.job_id = job.circuits_run_data[job_idx].results.job_id

        self.job_dir = (
            Path.home()
            / "Documents" / "QuantumInspireProjects" / self.date_timestamp
            / f"{self.job_0_timestamp}_{self.project_name}"
            / f"job_idx_{job_idx}__job_id_{self.job_id}"
        )
        self.job_dir.mkdir(parents=True, exist_ok=True)

    def store_job_result(self,
                         job: QIJob,
                         job_idx: int):
        """
        This instance method stores the job results (with the exception of
        the raw data), i.e. relevant job metadata and the result 'counts'.

        Args:
            job (QIJob):
                The user already-submitted job (project) object.

            job_idx (int):
                The job index for all jobs contained within the project.
                While a project may contain a certain number of jobs, e.g. N,
                it is generally true that the execution of all these jobs
                in the Quantum Inspire platform is not sequential with respect
                to the order with which those jobs were created.
                Therefore, job_idx is being utilized for clarity when storing
                the data, so that it follows the sequence with which the jobs were
                created.
        """
        
        self.result_id = job.circuits_run_data[job_idx].results.id
        self.execution_time_in_seconds = job.circuits_run_data[job_idx].results.execution_time_in_seconds
        self.shots_requested = job.circuits_run_data[job_idx].results.shots_requested
        self.shots_done = job.circuits_run_data[job_idx].results.shots_done
        if job.circuits_run_data[job_idx].results.raw_data == None:
            self.raw_data_memory = False
        else:
            self.raw_data_memory = True

        self.counts = job.circuits_run_data[job_idx].results.results

        job_result_dict = {}
        job_result_dict['Job timestamp'] = f"{self.date_timestamp}_{self.job_timestamp}"
        job_result_dict['Job ID'] = self.job_id
        job_result_dict['Result ID'] = self.result_id
        job_result_dict['Circuit name'] = self.circuit_name
        job_result_dict['Number of qubits specified'] = self.num_qubits
        job_result_dict['Number of classical bits specified'] = self.num_clbits
        job_result_dict['Circuit depth'] = self.circuit_depth
        job_result_dict['Execution time in seconds'] = self.execution_time_in_seconds
        job_result_dict['Shots requested'] = self.shots_requested
        job_result_dict['Shots done'] = self.shots_done
        job_result_dict['Raw data memory'] = self.raw_data_memory
        job_result_dict['Counts'] = self.counts
        file_path = (
            Path(self.job_dir)
            / f"job_result_{self.date_timestamp}_{self.job_timestamp}.json"
        )
        with open(file_path, 'w') as file:
            json.dump(job_result_dict, file, indent=3)

    def store_circuit_metadata(self,
                               job: QIJob,
                               job_idx: int):
        """
        This instance method stores the job circuit metadata, i.e. bookkeeping
        records, cQASM_v3 and OpenQASM3 program files, as well as the circuit
        figure.

        Args:
            job (QIJob):
                The user already-submitted job (project) object.

            job_idx (int):
                The job index for all jobs contained within the project.
                While a project may contain a certain number of jobs, e.g. N,
                it is generally true that the execution of all these jobs
                in the Quantum Inspire platform is not sequential with respect
                to the order with which those jobs were created.
                Therefore, job_idx is being utilized for clarity when storing
                the data, so that it follows the sequence with which the jobs were
                created.
        """

        self.qc = job.circuits_run_data[job_idx].circuit
        self.circuit_name = self.qc.name
        self.num_qubits = self.qc.to_instruction().num_qubits
        self.num_clbits = self.qc.to_instruction().num_clbits
        self.circuit_depth = self.qc.depth()

        qasm3_program = qasm3.dumps(self.qc)
        cqasm_v3_program = dumps(self.qc)
        qasm3_program_path = (
            Path(self.job_dir)
            / f"qasm3_program_{self.date_timestamp}_{self.job_timestamp}.qasm"
        )
        cqasm_v3_program_path = (
            Path(self.job_dir)
            / f"cqasm_v3_program_{self.date_timestamp}_{self.job_timestamp}.cq"
        )
        with open(qasm3_program_path, 'w') as f:
            f.write(qasm3_program)
        with open(cqasm_v3_program_path, 'w') as f:
            f.write(cqasm_v3_program)

        fig1 = self.qc.draw('mpl', scale=1.3)
        fig1.suptitle(f'\n{self.date_timestamp}_{self.job_timestamp}\nTranspiled quantum circuit\nCircuit name: {self.circuit_name}\nJob ID: {self.job_id}\n',
                      x = 0.5, y = 0.99, fontsize=16)
        fig1.supxlabel(f'Circuit depth: {self.circuit_depth}', x = 0.5, y = 0.06, fontsize=18)
        circuit_fig_path = (
            Path(self.job_dir)
            / f"quantum_circuit_{self.date_timestamp}_{self.job_timestamp}.png"
        )
        fig1.savefig(circuit_fig_path)

    def store_raw_data(self,
                       job: QIJob,
                       job_idx: int):
        """
        This instance method stores the job raw data results in an HDF5
        file format.

        It is important here to mention the conventions: assume that a quantum
        circuit contains M number of mid-circuit measurements, and that we execute
        it with N number of shots. The raw data (raw_data) is returned in the form
        of a list of bitstrings, where each entry of that list represents a circuit shot,
        and each bitstring contains all measurement outcomes from each mid-circuit
        measurement block.

        In the Tuna backends, the rightmost bit in the bitstring is the first measurement
        outcome, while the leftmost bit is the final measurement outcome,
        e.g. for a bit register of size 3, the bitstring '001' means that the first
        measurement outcome was '1' and was stored in the bit c0, while the second and
        third measurement outcomes were '0' and were stored in the bits c1 and c2.

        In the HDF5 file, we reverse the order of the bitstring for clarity. The file
        contains a 2D array of N rows, representing each shot, and M columns, representing
        all mid-circuit measurement outcomes. Column 0 in this case represents the first
        measurement outcome, while column M-1 represents the final measurement outcome
        for a particular measurement shot. 

        Args:
            job (QIJob):
                The user already-submitted job (project) object.

            job_idx (int):
                The job index for all jobs contained within the project.
                While a project may contain a certain number of jobs, e.g. N,
                it is generally true that the execution of all these jobs
                in the Quantum Inspire platform is not sequential with respect
                to the order with which those jobs were created.
                Therefore, job_idx is being utilized for clarity when storing
                the data, so that it follows the sequence with which the jobs were
                created.
        """

        raw_data = job.circuits_run_data[job_idx].results.raw_data
        job_raw_data = []

        for circuit_shot_idx in range(len(raw_data)):

            raw_data_row = [int(raw_data[circuit_shot_idx][digit_idx]) for digit_idx in range(len(raw_data[0]))]
            raw_data_row_reversed = raw_data_row[::-1] # because results are printed reversed
            job_raw_data.append(raw_data_row_reversed)

        job_raw_data = np.array(job_raw_data, dtype=np.int8)

        hdf5_file_dir = (
            Path(self.job_dir)
            / f"raw_data_{self.date_timestamp}_{self.job_timestamp}.hdf5"
        )
        with h5py.File(hdf5_file_dir, 'w') as file:
            file.create_dataset('Experimental Data/Data', data=job_raw_data, compression="gzip")

class RetrieveProjectRecord:
    """
    This class is responsible for retrieving a single job record
    within the user's local 'Documents' directory, by providing as inputs
    the appropriate job timestamp, name, and Job ID. Those three variables
    uniquely identify a single job in the projects directory.

    It is meant to mimic 'result = job.result()', so that when the user runs

    loaded_result = RetrieveProjectRecord(timestamp, project_name, job_id)

    then they can obtain the measurement counts and raw data as

    loaded_result.get_counts()
    loaded_result.get_memory()
    """

    def __init__(self,
                 timestamp: str,
                 project_name: str,
                 job_id: str):
        """
        Args:
            timestamp (str):
                The job complete timestamp, to be parsed as 'YYYYMMDD_HHMMSS'.
            
            project_name (str):
                The original project name.
            
            job_id (str):
                The Job ID, as this appears also in the Quantum Inspire platform.
        """
        
        self.timestamp = timestamp
        self.project_name = project_name
        self.job_id = job_id

        date_timestamp = self.timestamp.split('_')[0]
        job_timestamp = self.timestamp.split('_')[1]

        project_dir = (
            Path.home() / "Documents" / "QuantumInspireProjects" / date_timestamp
            / f"{job_timestamp}_{self.project_name}"
        )

        self.job_dir = None
        jobs_folders = [p for p in project_dir.iterdir() if p.is_dir()]
        for dir_entry in jobs_folders:
            if self.job_id in dir_entry.name:
                self.job_dir = dir_entry
        if self.job_dir == None:
            raise ValueError(f'No files found for timestamp: {timestamp}, project name: {project_name}, and Job ID: {job_id}')

        self.retrieve_qc()
        self.get_counts()
        self.get_memory()

    def retrieve_qc(self):
        """
        This instance method retrieves the QuantumCircuit object of the job.
        """

        qasm3_file_path = (
            Path(self.job_dir)
            / f"qasm3_program_{self.timestamp}.qasm"
        )
        self.qc = qasm3.load(qasm3_file_path)

    def get_counts(self):
        """
        This instance method retrieves the job counts in a dictionary format.
        """

        json_file_path = (
            Path(self.job_dir)
            / f"job_result_{self.timestamp}.json"
        )

        with open(json_file_path, 'r') as file:
            json_data = json.load(file)

        counts = json_data['Counts']
        return counts

    def get_memory(self,
                   dummy_circuit_nr: int = None):
        """
        This instance method retrieves the job raw data in a list of string
        bitstrings format.

        For jobs were the variable 'memory' was set to False,
        e.g. job = backend.run(qc, shots=nr_shots, memory = False)
        this instance method will return an empty list.

        Args:
            dummy_circuit_nr (int):
                This is a dummy variable which does not affect the instance method.
                It exists so that the instance method mimics result = job.result().get_memory(),
                which does have the 'circuit_nr' as an input.
                In result = job.result().get_memory(circuit_nr), since the job object in
                reality is a project and can in principle contain multiple jobs, the variable
                circuit_nr is used to identify uniquely a single circuit tied to a single job.
                
                Since when instantiating the RetrieveProjectRecord class we have already
                identified a single job, the circuit_nr would have no meaning.
                Still, we use here a dummy circuit_nr variable for compatibility purposes with
                respect to other functions used in other modules.
        """

        hdf5_file_dir = (
            Path(self.job_dir)
            / f"raw_data_{self.timestamp}.hdf5"
        )

        try:
            with h5py.File(hdf5_file_dir, "r") as f:
                hdf5_data = f["Experimental Data"]["Data"][()]
            raw_shots = []
            for shot_idx in range(len(hdf5_data)):
                shot_string = ''.join(map(str, hdf5_data[shot_idx][::-1]))
                raw_shots.append(shot_string)
            return raw_shots

        except:
            return []


"""
Utility functions for generic quantum information calculations.

Authors: Marios Samiotis
"""

from qiskit.quantum_info import DensityMatrix, Pauli

def calculate_observable_value(density_state: DensityMatrix,
                               observable: str):
    """
    This function calculates the expectation value of an observable
    given an input quantum state (expressed in the density matrix representation).

    Args:
        density_state (DensityMatrix):
            The input quantum state, given in the density matrix representation.

        observable (str):
            The observable for which this function calculates the expectation values of.

            For an n-qubit Pauli string P, where Pauli Pi acts on qubit qi, the order
            in the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit observable string 'YX', qubit q0 is measured
            in the X basis, while qubit q1 is measured in the Y basis.
    """
    
    observable_operator = Pauli(observable)
    observable_expectation_value = density_state.expectation_value(observable_operator)

    return observable_expectation_value


"""
Utility functions for processing the raw data shots that have been obtained
from circuits that may also contain mid-circuit measurements.

NOTE: for all of these functions to be useful, the user must request the memory
of a job prior to executing it,
e.g. job = backend.run(qc, shots, memory = True)

Authors: Marios Samiotis
"""

import numpy as np
from qiskit import QuantumCircuit
from qiskit.result.result import Result
from qiskit.quantum_info import SparsePauliOp

def obtain_binary_list(nr_qubits: int):
    """
    This function returns an ordered list of binary numbers as a function
    of nr_qubits.

    e.g. for nr_qubits = 2, binary_list = ['00', '01', '10', '11']

    Args:
        nr_qubits (int):
            The total number of qubits for which the binary list will be
            created for.
    """

    binary_list = []
    for binary_str_idx in range(2**nr_qubits):
        binary_list.append(np.binary_repr(binary_str_idx, nr_qubits))
    return binary_list

def get_raw_data(qc: QuantumCircuit,
                 result: Result,
                 circuit_nr: int = None):
    """
    This function returns the raw_data for each measurement block within
    a given quantum circuit 'qc' in an organized way.

    Args:
        qc (QuantumCircuit):
            The quantum circuit object.

        result (Result):
            The result of a job (project), as returned from
            result = job.result()

        circuit_nr (int):
            The circuit number within a job, since a job can contain
            multiple quantum circuits.
            Defaults to None for a job with a single quantum circuit.
    """

    bit_register_size = qc.num_clbits
    raw_data = result.get_memory(circuit_nr)
    for entry in range(len(raw_data)):
        additional_len = bit_register_size - len(raw_data[entry])
        for i in range(additional_len):
            raw_data[entry] = '0' + raw_data[entry]

    return raw_data

def get_multi_counts(raw_data_shots: list,
                     nr_qubits: int):
    """
    This function returns a list containing entries of all count dictionaries
    for each measurement block within one quantum circuit.
    e.g. total_counts[0] contains the counts dictionary for the very first
    measurement block of a given quantum circuit.

    Args:
        raw_data_shots (list):
            The raw data shots returned from quantum circuits containing
            a single or multiple mid-circuit measurement blocks.

            For a quantum circuit containing M number of mid-circuit measurements,
            and executing it for a number of N shots, len(raw_data_shots) = N,
            while each entry of the list contains a bitstring of size M.

            The convention followed in each bitstring for a bit register of size K
            is 'cK-1,cK-2,...,c2,c1,c0', meaning that the rightmost bit corresponds
            to the very first bit in the bit register.

        nr_qubits (int):
            The number of qubits of the original quantum circuit.
    """

    binary_list = obtain_binary_list(nr_qubits)
    mid_circuit_blocks_nr = int(len(raw_data_shots[0]) / nr_qubits)

    total_counts = []
    for mcm_block_idx in range(mid_circuit_blocks_nr):

        counts_dict = {}
        for entry in binary_list:
            counts_dict[entry] = 0

        for shot_idx in range(len(raw_data_shots)):
            reversed_shots = raw_data_shots[shot_idx][::-1]

            binary_string_reversed = reversed_shots[mcm_block_idx*nr_qubits:(mcm_block_idx+1)*nr_qubits]
            # in order to ensure that len(binary_string) == nr_qubits,
            binary_string = binary_string_reversed[::-1]
            binary_string = np.binary_repr(int(binary_string, 2), nr_qubits)
            counts_dict[binary_string] += 1

        total_counts.append(counts_dict)
    return total_counts

def get_multi_probs(raw_data_counts: list[dict]):
    """
    This function takes in a list of the total measurement counts for each
    measurement block in a quantum circuit, and computes
    the probabilities for each block.

    Args:
        raw_data_counts (list[dict]):
            The total measurement counts for each (mid-circuit) measurement
            block in a given quantum circuit. The first entry of the list
            corresponds to the measurement counts of the first measurement
            block, while the last entry corresponds to the counts of the
            last measurement block.
    """

    nr_shots = 0
    for entry in raw_data_counts[0]:
        nr_shots += raw_data_counts[0][entry]

    probabilities = []
    for entry_idx in range(len(raw_data_counts)):
        prob_dict = {}
        for entry in raw_data_counts[entry_idx]:
            prob_dict[entry] = raw_data_counts[entry_idx][entry] / nr_shots
        probabilities.append(prob_dict)

    return probabilities


def observable_expectation_values_Z_basis(probabilities: list[dict],
                                          observable: str):
    """
    This function calculates the expectation values of an observable
    in the Z basis, given a list of measurement probabilities.
    In case the user wants to calculate the expectation value of an observable
    containing Paulis X and Y, first the appropriate pre-measurement rotations
    must be applied in the quantum circuit so that those are projected to the
    Z basis.

    Args:
        probabilitites (list[dict]):
            A list of dictionaries each containing the measurement probabilities
            of a certain measurement block.
            e.g. probabilitities = [{'00': 0.25, '01': 0.25, '10': 0.25, '11': 0.25},
                                    ...]

        observable (str):
            The observable for which this function calculates the expectation values of.
            It must be expressed in the Z basis, containing no Paulis X or Y.

            For an n-qubit Pauli string P, where Pauli Pi acts on qubit qi, the order
            in the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit observable string 'IZ', operator Z corresponds to
            qubit q0 while operator I corresponds to qubit q1.
    """

    if 'X' in observable or 'Y' in observable:
        raise ValueError(f"Observable {observable} must not contain Paulis X or Y.")
    
    nr_qubits = len(observable)
    binary_list = obtain_binary_list(nr_qubits)

    observable_values = []
    observable_matrix = np.real(SparsePauliOp([observable]).to_matrix())

    for entry_idx in range(len(probabilities)):
        observable = 0
        for binary_idx in range(len(binary_list)):
            observable += probabilities[entry_idx][binary_list[binary_idx]] * observable_matrix[binary_idx][binary_idx]
        observable_values.append(observable)

    return observable_values


"""
Utility functions for synthesizing a Trotterization quantum circuit.

Authors: Marios Samiotis
"""

import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit.library import PauliEvolutionGate
from qiskit.quantum_info import SparsePauliOp
from qi_utilities.utility_functions.circuit_modifiers import apply_pre_measurement_rotations

def apply_trotter_block(qc: QuantumCircuit,
                        hamiltonian: SparsePauliOp,
                        trotter_order: int,
                        time_step: float):
    """
    This function applies a Trotter block on a given quantum circuit
    'qc'.
    We follow the common convention hbar=1.

    Args:
        qc (QuantumCircuit):
            The quantum circuit object.

        hamiltonian (SparsePauliOp):
            The Hamiltonian operator describing the dynamics of a given
            quantum system, written in the Pauli basis.
            It should be given in units of [Hz].

            In general, a Hamiltonian operator is written as
            H = sum_{j} a_j * P_j, where a_j are complex coefficients in units of [Hz],
            and P_j are n-qubit Pauli operators (else referred to as 'Pauli strings').

            For an n-qubit Pauli operator P in the Hamiltonian, with Pauli Pi acting
            on qubit qi, the ordering of the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit Pauli string 'YX', operator X corresponds
            to qubit q0, while operator Y corresponds to qubit q1.

        trotter_order (int):
            The trotterization order. For more info, visit
            https://en.wikipedia.org/wiki/Hamiltonian_simulation#Product_formulas

        time_step (float):
            The time step in which we evolve dynamically forward in time
            the quantum state encoded in the quantum circuit 'qc'.
    """
    hamiltonian = hamiltonian[::-1]
    for entry in range(len(hamiltonian)):
        pauli_string = hamiltonian[entry].paulis.to_labels()[0]

        string_to_list = list(pauli_string)
        target_qubits = []
        qubit_counter = 0
        for index in reversed(range(len(string_to_list))):
            if string_to_list[index] == 'I':
                pass
            else:
                target_qubits.append(qubit_counter)
            qubit_counter += 1

        unitary_label = 'Trotter block,' + f' Pauli: {pauli_string}' + \
                        f'\nn = {trotter_order},' + f' Time = {time_step*1e9:.2f} ns'
        unitary_gate = PauliEvolutionGate(operator = hamiltonian[entry],
                                          time = time_step / trotter_order,
                                          label = unitary_label)
        qc.append(unitary_gate, target_qubits)

def construct_trotterization_circuit(initial_state: str,
                                     measured_observable: str,
                                     hamiltonian: SparsePauliOp,
                                     trotter_order: int,
                                     evolution_times: np.ndarray,
                                     time_step: float,
                                     midcircuit_measurement: bool = False):
    """
    This function uses the 'trotter_block' function contained in this module
    to create the Trotterization quantum circuit from an initial state.
    We follow the common convention hbar=1.

    Args:
        initial_state (str):
            A bitstring specifying the initial state. The order in the bitstring
            is 'qn-1,qn-2,...,q2,q1,q0'. The string can only contain '0' or '1'.
            e.g. for a quantum circuit with 3 qubits, initial_state = '100' will
            initialize qubits q0 and q1 in the state |0>, and qubit q2 in the state
            |1>.

        measured_observable (str):
            The observable whose expectation value we estimate at the end of
            the Trotterization algorithm.

            For an n-qubit Pauli string P, where Pauli Pi acts on qubit qi, the order
            in the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit observable string 'YX', qubit q0 is measured
            in the X basis, while qubit q1 is measured in the Y basis.

        hamiltonian (SparsePauliOp):
            The Hamiltonian operator describing the dynamics of a given
            quantum system, written in the Pauli basis.
            It should be given in units of [Hz].

            In general, a Hamiltonian operator is written as
            H = sum_{j} a_j * P_j, where a_j are complex coefficients in units of [Hz],
            and P_j are n-qubit Pauli operators (else referred to as 'Pauli strings').

            For an n-qubit Pauli operator P in the Hamiltonian, with Pauli Pi acting
            on qubit qi, the ordering of the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit Pauli string 'YX', operator X corresponds
            to qubit q0, while operator Y corresponds to qubit q1.

        trotter_order (int):
            The trotterization order. For more info, visit
            https://en.wikipedia.org/wiki/Hamiltonian_simulation#Product_formulas

        evolution_times (np.ndarray):
            A numpy array containing the discrete time steps for which the simulation
            solves the time-dependent Schrödinger equation.

        time_step (float):
            The time step in which we evolve dynamically forward in time
            the quantum state encoded in the quantum circuit 'qc'.

        midcircuit_measurement (bool):
            A flag which should be set to 'True' if we are utilizing the mid-circuit
            functionality when constructing the Trotterization algorithm.
    """

    nr_qubits = len(initial_state)
    if midcircuit_measurement == True:
        qc = QuantumCircuit(nr_qubits, nr_qubits*len(evolution_times))
    else:
        qc = QuantumCircuit(nr_qubits, nr_qubits)
    for idx in range(nr_qubits):
        qc.reset(idx)
    qc.barrier()

    for idx in range(len(initial_state)):
        if initial_state[idx] == '1':
            qc.x((nr_qubits-1) - idx)
    qc.barrier()

    for repetition in range(1, trotter_order+1):
        apply_trotter_block(qc, hamiltonian, trotter_order, evolution_times[time_step])
    qc.barrier()

    if midcircuit_measurement == True:
        apply_pre_measurement_rotations(qc, measured_observable, [nr_qubits*time_step, nr_qubits*time_step + 1])
    else:
        apply_pre_measurement_rotations(qc, measured_observable)
    qc.barrier()

    return qc


"""
Utility functions for solving classically the time-dependent
Schrödinger equation, given an initial quantum state and a system
Hamiltonian.

Authors: Marios Samiotis
"""

import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit.library import PauliEvolutionGate
from qiskit.quantum_info import SparsePauliOp, DensityMatrix, Operator
from qi_utilities.utility_functions.quantum_info import calculate_observable_value

def evolve_quantum_state(quantum_state: QuantumCircuit,
                         hamiltonian: SparsePauliOp,
                         time_step: float):
    """
    This function evolves a (initial) quantum state by applying
    the time evolution unitary operator U(t) = exp(-iHt) on the
    state, for a time_step 't' and a Hamiltonian operator 'H'.

    The initial state 'quantum_state' must be a quantum circuit.
    We follow the common convention hbar=1.

    Args:
        quantum_state (QuantumCircuit):
            The quantum state to be evolved, given as a QuantumCircuit object.
            The user is required to create such a circuit outside of this function.

        hamiltonian (SparsePauliOp):
            The Hamiltonian operator describing the dynamics of a given
            quantum system, written in the Pauli basis.
            It should be given in units of [Hz].

            In general, a Hamiltonian operator is written as
            H = sum_{j} a_j * P_j, where a_j are complex coefficients in units of [Hz],
            and P_j are n-qubit Pauli operators (else referred to as 'Pauli strings').

            For an n-qubit Pauli operator P in the Hamiltonian, with Pauli Pi acting
            on qubit qi, the ordering of the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit Pauli string 'YX', operator X corresponds
            to qubit q0, while operator Y corresponds to qubit q1.

        time_step (float):
            The time step in which we evolve dynamically forward in time
            the quantum_state.
    """

    if type(quantum_state) is not QuantumCircuit:
        raise TypeError(f"Object {quantum_state} must be a QuantumCircuit.")
    
    density_state = DensityMatrix(quantum_state)
    evolution_matrix = PauliEvolutionGate(operator = hamiltonian,
                                          time = time_step)
    evolution_operator = Operator(evolution_matrix)
    evolved_state = density_state.evolve(evolution_operator)

    return evolved_state

def simulate_time_evolution(initial_state: QuantumCircuit,
                            hamiltonian: SparsePauliOp,
                            evolution_times: np.ndarray,
                            observables: list):
    """
    This function takes as inputs an initial quantum state, the system Hamiltonian,
    the evolution times and the list of observables to be extracted, and solves
    the time-dependent Schrödinger equation.
    It returns a dictionary containing the values of each observable in the
    'observables' list, for all time steps in 'evolution_times'.

    The initial state 'initial_state' must be a quantum circuit.
    We follow the common convention hbar=1.

    Args:
        initial_state (QuantumCircuit):
            The quantum state to be evolved, given as a QuantumCircuit object.
            The user is required to create such a circuit outside of this function.

        hamiltonian (SparsePauliOp):
            The Hamiltonian operator describing the dynamics of a given
            quantum system, written in the Pauli basis.
            It should be given in units of [Hz].

            In general, a Hamiltonian operator is written as
            H = sum_{j} a_j * P_j, where a_j are complex coefficients in units of [Hz],
            and P_j are n-qubit Pauli operators (else referred to as 'Pauli strings').

            For an n-qubit Pauli operator P in the Hamiltonian, with Pauli Pi acting
            on qubit qi, the ordering of the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit Pauli string 'YX', operator X corresponds
            to qubit q0, while operator Y corresponds to qubit q1.

        evolution_times (np.ndarray):
            A numpy array containing the discrete time steps for which the simulation
            solves the time-dependent Schrödinger equation.

        observables (list):
            A list containing the observables for which the simulation calculates
            the expectation values of, for each time step within the 'evolution_times'.

            For an n-qubit Pauli string P, where Pauli Pi acts on qubit qi, the order
            in the string is 'Pn-1,Pn-2,...,P2,P1,P0'.
            e.g. for the two-qubit observable string 'IZ', operator Z corresponds to
            qubit q0 while operator I corresponds to qubit q1.
    """

    if type(initial_state) is not QuantumCircuit:
        raise TypeError(f"Object {initial_state} must be a QuantumCircuit.")
    
    dt = evolution_times[1] - evolution_times[0] #use this for the time_interval in noisy evolution
    observables_dict = {}
    for observable in observables:
        observables_dict[observable] = {}
        observables_dict[observable]['list'] = list(observable)
        observables_dict[observable]['values'] = []
    for time_step in evolution_times:
        evolved_state = evolve_quantum_state(initial_state, hamiltonian, time_step)
        # put here the evolution due to a noise channel
        for observable in observables:
            observable_value = calculate_observable_value(evolved_state, observable)
            observables_dict[observable]['values'].append(observable_value)

    return observables_dict


"""
Utility functions for processing raw measurement data for executed
quantum circuits which contained measurements for extracting the readout
assignment matrix, and for applying the readout assignment matrix to
raw data shots in order to mitigate readout errors.

Authors: Marios Samiotis
"""

import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
from matplotlib.colors import LinearSegmentedColormap, Normalize
from qiskit.result.result import Result
from qi_utilities.utility_functions.raw_data_processing import obtain_binary_list, get_multi_counts

def split_raw_shots(result: Result,
                    qubit_list: list,
                    circuit_nr: int = None):
    """
    This function splits the raw data shots of the measurement result
    into two groups, for a circuit containing readout mitigation circuits
    at its end.

    Args:
        result (Result):
            The result of a job (project), as returned from
            result = job.result()

        qubit_list (list):
            An ordered list containing the integer indices of the qubits
            used in the original quantum circuit.
            e.g. for qubits q0 and q2, qubit_list = [0, 2].

        circuit_nr (int):
            The circuit number within a job, since a job can contain
            multiple quantum circuits.
            Defaults to None for a job with a single quantum circuit.
    """

    nr_qubits = len(qubit_list)
    raw_shots = result.get_memory(circuit_nr)

    experiment_shots = []
    ro_mitigation_shots = []

    for raw_shots_entry in range(len(raw_shots)):
        ro_mitigation_shots.append(raw_shots[raw_shots_entry][0:nr_qubits*2**nr_qubits])
        experiment_shots.append(raw_shots[raw_shots_entry][nr_qubits*2**nr_qubits::])
    return experiment_shots, ro_mitigation_shots

def get_ro_corrected_multi_probs(raw_data_probs: list[dict],
                                 ro_assignment_matrix: np.ndarray,
                                 qubit_list: list):
    """
    This function takes as input the raw data probabilities and applies the
    readout assignment matrix in order to mitigate readout errors.

    Args:
        raw_data_probs (list):
            A list of dictionaries each containing the measurement probabilities
            of a certain measurement block.
            e.g. raw_data_probs = [{'00': 0.25, '01': 0.25, '10': 0.25, '11': 0.25},
                                    ...]

        ro_assignment_matrix (np.ndarray):
            The readout assignment matrix.

        qubit_list (list):
            An ordered list containing the integer indices of the qubits
            used in the original quantum circuit.
            e.g. for qubits q0 and q2, qubit_list = [0, 2].
    """

    nr_qubits = len(qubit_list)
    binary_list = obtain_binary_list(nr_qubits)
    
    raw_data_probs_ro_corrected = []

    for entry_idx in range(len(raw_data_probs)):

        probs = np.array(list(raw_data_probs[entry_idx].values()))

        def objective(x):
            return np.linalg.norm(ro_assignment_matrix @ x - probs) ** 2
        constraints = {
            "type": "eq",
            "fun": lambda x: np.sum(x) - 1
        }
        bounds = [(0, 1)] * len(probs)
        result = minimize(
            objective,
            x0=np.ones(len(probs)) / len(probs),  # initial guess: uniform distribution
            method="SLSQP",
            bounds=bounds,
            constraints=constraints
        )

        probs_ro_corrected_dict = {}
        for idx in range(len(result.x)):
            probs_ro_corrected_dict[binary_list[idx]] = result.x[idx]

        raw_data_probs_ro_corrected.append(probs_ro_corrected_dict)
    return raw_data_probs_ro_corrected


def extract_ro_assignment_matrix(ro_mitigation_shots: list,
                                 qubit_list: list):
    """
    This function uses as input the raw data shots from the readout
    mitigation circuits that are usually appended at the end of a
    quantum circuit, and extracts the readout assignment matrix.

    Args:
        ro_mitigation_shots (list):
            The raw data shots returned from the readout mitigation
            quantum circuits which contain multiple mid-circuit
            measurement blocks.

            For a quantum circuit containing M number of mid-circuit measurements,
            and executing it for a number of N shots, len(raw_data_shots) = N,
            while each entry of the list contains a bitstring of size M.

            The convention followed in each bitstring for a bit register of size K
            is 'cK-1,cK-2,...,c2,c1,c0', meaning that the rightmost bit corresponds
            to the very first bit in the bit register.

        qubit_list (list):
            An ordered list containing the integer indices of the qubits
            used in the original quantum circuit.
            e.g. for qubits q0 and q2, qubit_list = [0, 2].
    """

    nr_qubits = len(qubit_list)
    prepared_states = obtain_binary_list(nr_qubits)
    
    ro_counts_per_prepared_states = get_multi_counts(ro_mitigation_shots, nr_qubits)
    assignment_probability_matrix = np.zeros([len(prepared_states), len(prepared_states)], dtype=np.float64)
    for prepared_state_idx in range(len(prepared_states)):
        assignment_probability_matrix[prepared_state_idx] = np.array([ro_counts_per_prepared_states[prepared_state_idx][state]
                                                 for state in prepared_states]) / len(ro_mitigation_shots)

    return assignment_probability_matrix


def plot_ro_assignment_matrix(ro_assignment_matrix: np.ndarray,
                              qubit_list: list):
    """
    This function creates and prints a figure of the readout assignment
    matrix.

    Args:
        ro_assignment_matrix (np.ndarray):
            The readout assignment matrix.

        qubit_list (list):
            An ordered list containing the integer indices of the qubits
            used in the original quantum circuit.
            e.g. for qubits q0 and q2, qubit_list = [0, 2].
    """

    def red_white_green_cmap():
        n = 256

        # Fractions of the full range
        reds_frac   = 20 / 100
        middle_frac = 50 / 100 - reds_frac

        reds_n   = int(n * reds_frac)
        middle_n = int(n * middle_frac)
        greens_n = n - reds_n - middle_n

        reds = plt.cm.Reds(np.linspace(0.0, 1.0, reds_n))
        middle = np.ones((middle_n, 4))  # white color
        greens = plt.cm.Greens(np.linspace(0.0, 1.0, greens_n))

        colors = np.vstack((reds, middle, greens))
        return LinearSegmentedColormap.from_list("RedWhiteGreen", colors)

    nr_qubits = len(qubit_list)
    binary_labels = []
    for binary_str_idx in range(2**nr_qubits):
        binary_labels.append(r"$|$" + f"{np.binary_repr(binary_str_idx, nr_qubits)}" + r"$\rangle$")

    fig_size = max(6, len(binary_labels) * 0.4)
    fig, ax = plt.subplots(figsize=(fig_size, fig_size), dpi=300)

    qubit_list_label = r"$|$"
    for qubit_idx in qubit_list[::-1]:
        qubit_list_label += f" Q{qubit_idx} "
    qubit_list_label += r"$\rangle$"

    ax.set_xticks(np.arange(len(binary_labels)))
    ax.set_yticks(np.arange(len(binary_labels)))
    ax.set_xticklabels(binary_labels)
    ax.set_yticklabels(binary_labels)

    base_fontsize = max(3, 14 - 0.2 * len(binary_labels))
    ax.tick_params(axis="x", labelsize=base_fontsize)
    ax.tick_params(axis="y", labelsize=base_fontsize)
    ax.tick_params(axis='x', rotation=60)

    ax.xaxis.tick_top()
    ax.set_xlabel("Declared state")
    ax.xaxis.set_label_position('top')
    ax.set_ylabel("Prepared state")
    ax.set_title(f"Readout assignment matrix\nQubit list: {qubit_list_label}")

    plt.setp(ax.get_xticklabels(), ha="center")

    values = np.zeros((len(binary_labels), len(binary_labels)))
    for i in range(len(binary_labels)):
        for j in range(len(binary_labels)):
            values[i, j] = ro_assignment_matrix[i, j] * 100
            cell_fontsize = max(2, 10 - 0.15 * len(binary_labels))
            txt = ax.text(j, i, f"{values[i, j]:.1f}%", ha="center", va="center",
                    color="white", fontweight="bold", fontsize=cell_fontsize)
            txt.set_path_effects([
                path_effects.Stroke(linewidth=2, foreground="black"),
                path_effects.Normal()
            ])

    cmap = red_white_green_cmap()
    norm = Normalize(vmin=0, vmax=100)

    cax = ax.imshow(values, cmap=cmap, norm=norm)
    cbar = fig.colorbar(cax, ax=ax, fraction=0.046, pad=0.04, label="Probability (%)")
    cbar.set_ticks(ticks=list(np.arange(0, 110, 10)))

    plt.tight_layout()
    plt.show()